#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Florian Briegel (briegel@mpia.de)
# @Date: 2021-08-12
# @Filename: jsonstring.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

'''
'''

from __future__ import annotations

import click
from clu.command import Command
from clu.parsers.click import command_parser

import json

class JsonStringParamType(click.ParamType):
    name = "jsonstring"

    def convert(self, value, param, ctx):
        try:
            return json.loads(value)
        
        except ValueError:
            self.fail(f"{value!r} is not a valid json string", param, ctx)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Florian Briegel (briegel@mpia.de)
# @Date: 2021-08-12
# @Filename: jsonpickle.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

'''
'''

from __future__ import annotations

import click
from clu.command import Command
from clu.parsers.click import command_parser

import jsonpickle
import numpy as np


class JsonPickleParamType(click.ParamType):
    name = "jsonpickle"

    def convert(self, value, param, ctx):
        try:
            return jsonpickle.decode(value)
        except ValueError:
            self.fail(f"{value!r} is not a valid jsonpickle", param, ctx)


def pickle(*argv):
    '''
    converts single or multiple data to a quoted json string.
    '''
    if(len(argv) > 1):
        return ["'" + jsonpickle.encode(a, make_refs=False) + "'" for a in argv]
    else:
        return "'" + jsonpickle.encode(argv[0], make_refs=False) + "'"

def unpickle(*argv):
    '''
    converts single or multiple data to a quoted json string.
    '''
    if(len(argv) > 1):
        return [jsonpickle.decode(a[1:-1]) for a in argv]
    else:
        return jsonpickle.decode(argv[0][1:-1])








# -*- coding: utf-8 -*-
#
# @Author: Florian Briegel (briegel@mpia.de)
# @Date: 2021-08-18
# @Filename: proxy.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


from __future__ import annotations

import asyncio
# import logging
import sys
from functools import partial
from itertools import chain
from typing import Callable, Optional
from shlex import quote
import json

from inspect import getcoroutinelocals, iscoroutine
from clu import AMQPReply, BaseClient

from .exceptions import ProxyPartialInvokeException



class Proxy():
    """A proxy client with actor commands.

import uuid
from clu import AMQPClient, CommandStatus
from cluplus.proxy import Proxy, invoke, unpack

actor = "proto"

amqpc = AMQPClient(name=f"proxy-{uuid.uuid4().hex[:8]}")

proto = Proxy(amqpc, actor)
proto.start()

async def start_now(proto):
    await amqpc.start()
    await proto.start()
    await amqpc.stop()

amqpc.loop.run_until_complete(start_now(proto))

print(proto.ping())

async def start_ping(proto):
    await amqpc.start()
    print(await proto.ping())
    await amqpc.stop()

amqpc.loop.run_until_complete(start_ping(proto))

async def start_async_ping(proto):
    await amqpc.start()
    print(await proto.async_ping())
    await amqpc.stop()

amqpc.loop.run_until_complete(start_async_ping(proto))


invoke(proto.async_setEnabled(True, axis0=True),
       proto.async_gotoRaDecJ2000(10,20))


unpack(proto.ping())

a, a0, b0, *coord = unpack(invoke(proto.async_setEnabled(True, axis0=True),
                                  proto.async_gotoRaDecJ2000(10,20)))
try:
   proto.errPassAsError()
   
except Exception as ex:
   print(f"got: {ex}")

try:
   amqpc.loop.run_until_complete(invoke(proto.ping(async_mode=True),
                                        proto.async_version(),
                                        proto.async_errPassAsError()))
except Exception as ex:
   print(ex)

try:
    async def start_async_invoke():
        await amqpc.start()
        await invoke(proto.ping(),
                     proto.version(),
                     proto.errPassAsError())
        await amqpc.stop()
    amqpc.loop.run_until_complete(start_async_invoke())
except Exception as ex:
   print(ex)


def callback(reply): 
    amqpc.log.warning(f"Reply: {CommandStatus.code_to_status(reply.message_code)} {reply.body}")

proto.setEnabled(True, callback=callback)


async def start_async_setEnabled():
        await amqpc.start()
        await proto.setEnabled(True, callback=callback)
        await amqpc.stop()

amqpc.loop.run_until_complete(start_async_setEnabled())
    """
    
    __commands = "__commands"
    __comkey = "help"
    
    def __init__(self, client: BaseClient, actor: str):
        self.client = client
        self.actor = actor
        self.async_mode = True
#        super().__init__(client, actor)

    def start(self):
        """Query and set actor commands."""

        def set_commands(reply):
            commands = reply[Proxy.__comkey] if isinstance(reply, dict) else reply.help

            for c in commands:
                setattr(self, c, partial(self._hybrid_call_command, c))
                setattr(self, f"async_{c}", partial(self.call_command, c))
                setattr(self, f"nowait_{c}", partial(self.call_command, c, nowait=True))


        if self.client.loop.is_running():
            async def start_async():
                set_commands(await self.call_command(Proxy.__commands))
            return start_async()
        else:
            self.async_mode = False
            coro = self._sync_call_command(Proxy.__commands)
            set_commands(self.client.loop.run_until_complete(coro))
            self.client.loop.run_until_complete(self.client.stop())
            return self


    def isClientConnected(self):
        if not self.client.connection.connection:
            return False
        return not self.client.connection.connection.is_closed

    async def _sync_call_command(self,
                                 command,
                                 *args,
                                 callback: Optional[Callable[[AMQPReply], None]] = None,
                                 object_hook: Optional[Callable[[AMQPReply], None]] = None,
                                 **kwargs):

        await self.client.start()

        try:
            ret = await self.call_command(command,
                                        *args,
                                        callback=callback,
                                        object_hook=object_hook,
                                        **kwargs)
        except Exception as ex:
            await self.client.stop()
            raise ex

        await self.client.stop()
        return ret

    def _hybrid_call_command(self,
                             command,
                             *args,
                             nowait:Bool = False,
                             callback: Optional[Callable[[AMQPReply], None]] = None,
                             object_hook: Optional[Callable[[AMQPReply], None]] = None,
                             **kwargs):

        if self.async_mode:
            return self.call_command(command,
                                     *args,
                                     nowait=nowait,
                                     callback=callback,
                                     object_hook=object_hook,
                                     **kwargs)
        else:
            return self.client.loop.run_until_complete(
                self._sync_call_command(command,
                                        *args,
                                        callback=callback,
                                        object_hook=object_hook,
                                        **kwargs))

    async def call_command(self,
                           command: str,
                           *args,
                           callback: Optional[Callable[[AMQPReply], None]] = None,
                           nowait:Bool = False,
                           object_hook: Optional[Callable[[AMQPReply], None]] = None,
                           **kwargs):

        def encode(v):
            if isinstance(v, (int, float, bool)): return v
            elif isinstance(v, str): return v if v[0] in "'\"" and v[-1] in "'\"" else quote(v)
            return f"'{json.dumps(v)}'"

        args = [encode(v) for v in args] \
             + list(chain.from_iterable(('--' + k, encode(v))
                                        for k, v in kwargs.items()))

        fu = await self.client.send_command(self.actor,
                                                command,
                                                *args,
                                                callback=callback)
        
        if nowait: return fu

        ret = await fu

        if hasattr(ret, "status") and ret.status.did_fail:
            raise self._errorMapToException(ret.replies[-1].body['error'])

        return ret.replies[-1].body




    @staticmethod
    def _errorMapToException(em):
        if isinstance(em, dict):
            return Proxy._stringToException(em['exception_message'],
                                            em['exception_type'],
                                            em['exception_module'])
        return Exception(em)


    @staticmethod
    def _stringToException(sval, tn='Exception', mn='builtins'):
        try:
            module = sys.modules[mn] if mn in sys.modules \
                                     else __import__(mn, fromlist=tn)
            return getattr(module, tn)(sval)

        except AttributeError:
            return Exception(f'Unknown exception type {tn}-{sval}')

        except ImportError:
            return Exception(f'Unknown module type {mn}-{tn}:{sval}')


def invoke(*cmds):
    """invokes one or many commands in parallel

    On error it throws an exception if one of the commands fails as a dict
    with an exception and return values for every command.
    """
    async def invoke_now(proxy, *cmds, return_exceptions=False):
        if proxy and not proxy.isClientConnected():
            await proxy.client.start()
        
        ret = await asyncio.gather(*[asyncio.create_task(cmd) for cmd in cmds], 
                                   return_exceptions=True)

        if proxy:
            await proxy.client.stop()

        if not return_exceptions:
            for r in ret:
                if isinstance(r, Exception):
                    raise ProxyPartialInvokeException(*ret)
        return ret


    first_coro = cmds[0]
    assert(iscoroutine(first_coro))

    proxy = getcoroutinelocals(first_coro)['self']
    loop = proxy.client.loop
    
    if loop.is_running():
        return invoke_now(None, *cmds)
    else:
        return loop.run_until_complete(invoke_now(proxy, *cmds))


def unpack(ret, *keys):
    """ unpacks every parameter from the body of the finish reply or list of replies.

        Pythons list unpacking mechanism PEP3132 can be used to assign the value(s)
        
        Be warned if you dont use it the correct way, 
        whould also be a good place to check the reply body format with schemas.

        >>> a, b, c = [1, 2, 3]
        >>> a
        1

        >>> a = [1, 2, 3]
        >>> a
        [1, 2, 3]

        >>> a, b = [1, 2, 3]
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        ValueError: too many values to unpack (expected 2)

        >>> a, *b = [1, 2, 3]
        >>> a
        1
        >>> b
        [2, 3]

        Parameters
        ----------
        keys
        return only the parameters from keys
    """

    if len(ret) == 0:
        return
    
    if isinstance(ret, list):
        if len(keys) > 0:
            return [d[i] for d in ret for i in keys if i in d]
        else:
            return [val for d in ret for val in list(d.values())]

    if len(ret) == 1:
        return list(ret.values())[0]
    elif len(keys) > 1:
        return [ret[i] for i in keys]
    else:
        return list(ret.values())




# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python'}

packages = \
['cluplus', 'cluplus.parsers']

package_data = \
{'': ['*']}

install_requires = \
['click-default-group>=1.2.2,<2.0.0',
 'daemonocle>=1.1.1,<2.0.0',
 'jsonpickle==1.3',
 'sdss-access>=0.2.3',
 'sdss-clu>=1.2.0',
 'sdss-tree>=2.15.2',
 'sdsstools>=0.4.0']

entry_points = \
{'console_scripts': ['plotloop = utils.plotloop:main',
                     'proto = proto.__main__:proto',
                     'singleframe = cluplus.__main__:main']}

setup_kwargs = {
    'name': 'sdss-cluplus',
    'version': '0.0.18',
    'description': 'Additional functionality for sdss-clu',
    'long_description': '\nCLU+\n==========================================\n\n|py| |pypi| |Build Status| |docs| |Coverage Status|\n\n``CLU+`` adds some enhancements to sdss-clu\n\nFeatures\n--------\n- RPC/Corba/Zeroc Ice style actor communications\n- Asyncio & classic python usage\n- Complex data transfer with json\n- Alternative data handling with json-pickle, eg: numpy\n- Recursive config file loading\n\nInstallation\n------------\n\n``CLU+`` can be installed using ``pip`` as\n\n.. code-block:: console\n\n    pip install sdss-cluplus\n\nor from source\n\n.. code-block:: console\n\n    git clone https://github.com/sdss/cluplus\n    cd cluplus\n    pip install .\n\n\nNext, head to the `Getting started <https://github.com/sdss/cluplus/wiki>`__ section for more information about using clu+.\n\n\n.. |Build Status| image:: https://img.shields.io/github/workflow/status/sdss/cluplus/Test\n    :alt: Build Status\n    :target: https://github.com/sdss/cluplus/actions\n\n.. |Coverage Status| image:: https://codecov.io/gh/sdss/cluplus/branch/master/graph/badge.svg?token=i5SpR0OjLe\n    :alt: Coverage Status\n    :target: https://codecov.io/gh/sdss/cluplus\n\n.. |py| image:: https://img.shields.io/badge/python-3.7%20|%203.8%20|%203.9-blue\n    :alt: Python Versions\n    :target: https://docs.python.org/3/\n\n.. |docs| image:: https://readthedocs.org/projects/docs/badge/?version=latest\n    :alt: Documentation Status\n    :target: https://cluplus.readthedocs.io/en/latest/?badge=latest\n\n.. |pypi| image:: https://badge.fury.io/py/sdss-cluplus.svg\n    :alt: PyPI version\n    :target: https://badge.fury.io/py/sdss-cluplus\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n',
    'author': 'Florian Briegel',
    'author_email': 'briegel@mpia.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sdss/cluplus',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
